#!/usr/bin/perl -s
#-------------------------------------------------------------------------------
#        1         2         3         4         5         6         7         8
#2345678901234567890123456789012345678901234567890123456789012345678901234567890
#-------------------------------------------------------------------------------
$VERSION = "5.1";
$DATE =    "07 Jan 2022";
#-------------------------------------------------------------------------------
#
# 5.1 Modified 09 Jan 2022  (Pierre VIAU)
# - added option to specify annotations as number of bytes (in case no known device)
#
# 5.0 Modified 07 Jan 2022  (Pierre VIAU)
# - added second scrambling polynom (needed for Biomass)
#
#-------------------------------------------------------------------------------
#
# 4.3 Modified 13 Nov 2018  (Pierre VIAU)
# - added Sentinel-6
# - added decoding and reporting of CCSDS version and spacecraft ID for each frame
#
# 4.2 Modified 21 Jul 2016  (Pierre VIAU)
# - change terminology header/trailer to leading/trailing (to avoid confusion with actual headers)
# - optimized RS polynom hard-coded lines
#
# 4.1 Modified 29 Jun 2016  (Pierre VIAU)
# - corrected EarthCare configuration
# - added version and date to the output
#
# 4.0 Modified 23 Jun 2016  (Pierre VIAU)
# - converged into one the several hacked branches (Sentinels, S5P, Aeolus)
# - added option to select CCSDS header version
# - added options to select frame and RS block sizes
# - added option to configure the above by simply selecting the mission
#
#-------------------------------------------------------------------------------
#
# 3.3 Modified 13 Aug 2015  (Pierre VIAU)
# - added option to log sequence of VCIDs
#
# 3.2 Modified 5 Dec 2014  (Pierre VIAU)
# - added option to re-sync if sync marker found incorrect within the file
# - added option to log information in case of (re-)sync needed
# - forced display of sync markers to 8 digits always"
#
# 3.1 Modified 14 Nov 2013  (Pierre VIAU)
# - corrected bug: when in stat mode only (no output file), scrambled input option was ignored
# - always print VCID counter (was only when splitting output per VCID)
# - small optimization: calculate scrambling bytes for VCID only once (was at every frame)
#
# 3.0 Modified 28 Sep 2013  (Pierre VIAU)
# - added option to split output per VCID
# - a better aligned printed output (will be fine until 1M frames)
# - cleaner printed output (no info on output if none is produced)
#
#-------------------------------------------------------------------------------
#
# 2.1 Modified 10 Jul 2013  (Pierre VIAU)
# - added systematic list of VCID statistics
# - added option to NOT produce output file (to allow statistics only)
# - added instructions and one example for this
#
# 2.0 Modified 6 Apr 2013  (Pierre VIAU)
# - added possibility to process input VCDUs
# - added knowledge of VCDU annotations
# - added -in and -out options to define files content (CADU or VCDU)
# - modified annotation options, VCDU is a file type not a possible annotation (was dirty...)
# - improved instructions (more details)
#
#-------------------------------------------------------------------------------
#
# 1.1 Modified 23 Mar 2013  (Pierre VIAU)
# - added offset switch to allow skipping leading bytes
# - added report of trailing bytes (i.e. incomplete CADU) discarded if any
# - removed ALL conversions, everything is now byte-strings operations
#   (for performance and coherence reasons)
#
# 1.0 Created 15 Mar 2013  (Pierre VIAU)
#
#-------------------------------------------------------------------------------
#
# Script to transform CCSDS CADUs or VCDUs (loosely called frames here)
# using various options
#
# To this effect:
# - check cmd-line switches for options, and configure conversion process
# - read frame per frame; this assumes a CADU or VCDU fixed size, configured at the top
# - convert as indicated by options
# - write resulting frame to separate output file
# - if specified, split output to several files, 1 per VCID
#
# NB1: the output file name is derived from input file name,
#      appending (a) meaningful file extension(s) which depend(s) on the options
#
# USAGE: (but run the tool without parameters for definite up-to-date information)
# 
#   CADU-convert.pl <switches> <frames input file>
#
#   switches:
#	-in=xxx              input frames type
#	                     . xxx = CADU | VCDU
#	                     . if not set, =CADU is assumed by default
#	-out=yyy             output frames type
#	                     . yyy = CADU | VCDU | NONE
#	                     . if not set, output type will be same as input type
#	                     . =CADU cannot be set with -in=VCDU (no encoding function)
#	                     . =NONE no output file will be produced, only statistics if any
#	-in_scrambled        input frames are scrambled
#	                     . cannot be set with -in=VCDU (only CADUs are scrambled)
#	-out_scrambled       output frames will be scrambled
#	                     . if not set and -in_scrambled is, output will be unscrambled
#	                     . if set and -in_scrambled is not, output will be scrambled
#	                     . cannot be set with -out=VCDU (only CADUs are scrambled)
#	-in_annotated=zzz    input frames are annotated
#	                     . zzz = HRDFEP | HDR | DLR | nnn (number of bytes)
#	-out_annotated       output frames will remain annotated
#	                     . if not set and -in_annotated is, annotations will be removed
#	                     . cannot be set if -in_annotated is not (no annotation function)
#	                     . cannot be set if -in=CADU and -out=VCDU (different types)
#	-offset=ttt          skip bytes at beginning of file
#	                     . ttt = n | AUTO
#	                     . =n skip n bytes
#	                     . =AUTO search for first sync marker
#                        . =AUTO cannot be set with -in=VCDU (no sync marker)
#   -sat=sss             satellite => auto-configures -header, -frame_size, -rs_size
#	                     . ss = S1 | S2 | S3 | S5P | S6 | AE | EC | BM
#                        . =S1  configures -header=v2, -frame_size=1912, rs_size=128, rs_type=223
#                        . =S2  configures -header=v2, -frame_size=1912, rs_size=128, rs_type=223
#                        . =S3  configures -header=v2, -frame_size=1912, rs_size=128, rs_type=223
#                        . =S5P configures -header=v2, -frame_size=1115, rs_size=160, rs_type=223
#                        . =S6  configures -header=v2, -frame_size=1115, rs_size=160, rs_type=223
#                        . =AE  configures -header=v1, -frame_size=1115, rs_size=160, rs_type=223
#                        . =EC  configures -header=v1, -frame_size=1115, rs_size=160, rs_type=223
#                        . =BM  configures -header=v2, -frame_size=1912, rs_size=128, rs_type=239
#   -header=vv           CCSDS frame header version
#                        . vv = v1 | v2
#                        . is ignored if -sat is set
#	                     . if neither is set, =v2 is assumed by default
#                        . =v1 uses 10-bits SCID, 3-bits VCID (8 channels)
#                        . =v2 uses 8-bits SCID, 6-bits VCID (64 channels)
#   -frame_size=ffff     frame (VCDU) size in bytes
#                        . is ignored if -sat is set
#	                     . if neither is set, =1912 is assumed by default
#   -rs_size=rrr         Reed-Solomon block size in bytes
#                        . is ignored if -sat is set
#	                     . if neither is set, =128 is assumed by default
#	-rs_type=yyy         Reed-Solomon type, either (255,223) or (255,239)
#                        . yyy = 223 | 239
#                        . is ignored if -sat is set
#                        . if neither is set, =223 is assumed by default
#	-no_idle             discard idle frames (works if scrambled or un-scrambled)
#	-per_VCID            split output to 1 file per VCID
#   -keep_bad_sync       do not abort on bad sync marker, search for next sync marker
#   -log                 if sync errors, report in log file with VCID and counter
#   -log_VCID            report in log file the sequence of VCID and counter for each frame
#	-flip_sync           correct HDR LVDS bug: LSB storing of the sync marker
#	
# EXAMPLES (with trivial file names of course):
#	
#	CADU-convert.pl frames_file
#	. no effect (all switches off)
#	
#	CADU-convert.pl -out_scrambled CADUs_file
#	. scrambles CADUs to file CADUs_file.scrambled
#	
#	CADU-convert.pl -in_scrambled scrambled_CADUs_file
#	. unscrambles CADUs to file scrambled_CADUs_file.un-scrambled
#	
#	CADU-convert.pl -in_scrambled -in_annotated=HDR -no_idle HDR_scrambled_CADUs_file
#	. unscrambles, removes HDR annotations, discards idle frames
#	. and stores to file HDR_scrambled_CADUs_file.un-scrambled.not-annotated.no_idle
#	
#	CADU-convert.pl -in_scrambled -out_scrambled -no_idle scrambled_CADUs_file
#	. removes scrambled idle frames to file scrambled_CADUs_file.no-idle
#	
#	CADU-convert.pl -out_scrambled -in_annotated=HDR -out_annotated HDR_CADUs_file
#	. scrambles, keeps HDR annotations to file HDR_CADUs_file.scrambled
#	
#	CADU-convert.pl -in=VCDU -in_annotated=HRDFEP HRDFEP_VCDUs_file
#	. removes VCDU HDR annotations to file HRDFEP_VCDUs_file.not-annotated
#	
#	CADU-convert.pl -in=CADU -out=VCDU -in_scrambled -in_annotated=HDR -no_idle HDR_CADUs_file
#	. unscrambles CADUs, removes HDR annotations and converts to VCDUs
#	. and stores to file HDR_CADUs_file.un-scrambled.VCDU.not-annotated
#
#	CADU-convert.pl -stats -out=NONE -in_annotated=HDR -in_scrambled HDR_scrambled_CADUs_file
#	. unscrambles CADUs, removes HDR annotations
#	. no output file is generated, used only for listing VCID statistics
#	
#===============================================================================
# configuration section - touch if you know what you are doing

$B_SM 	    =    4;			# Sync Marker size in Bytes

$B_VCDU		= 1912;			# default VCDU size in Bytes (as for Sentinels)
$B_RS		=  128;			# default Reed Solomon size in Bytes (as for Sentinels)

# $B_ann	=     ;			# annotation size is set below depending on options
#$B_ann_HDR_CADU		= 20;	# HDR CADU annotation size in Bytes
$B_ann_HDR_CADU		= 17;	# quick hack for short frames - eg S6 - HDR CADU annotation size in Bytes
$B_ann_HDR_VCDU		= 16;	# HDR VCDU annotation size in Bytes
$B_ann_HRDFEP_CADU	= 10;	# HRDFEP CADU annotation size in Bytes
$B_ann_HRDFEP_VCDU	= 18;	# HRDFEP VCDU annotation size in Bytes

$B_search	= 1000;			# sync marker search block size in bytes; must be < $B_CADU size

$correct_SM = "\x1a\xcf\xfc\x1d";	# CADU sync marker in byte-string hex

# $idle_VCID = 	;			# idle frame value is set below depending on options
# $mask_VCID = 	;			# idle frame mask is set below depending on options

# end of configuration section - do NOT touch anything below
#===============================================================================

$|++;	# disable output file buffering (only for debugging)

#-------------------------------------------------------------------------------
# check command-line arguments
print "\nCADU-convert.pl v$VERSION ($DATE)\n";

if (($#ARGV < 0) or ($#ARGV > 0)) { 
	
	print "\nUSAGE:\n\n";
	print "\tCADU-convert.pl <switches> <frames input file>\n\n";

	print "\tswitches:\n";
	print "\t-in=xxx              input frames type\n";
	print "\t                     . xxx = CADU | VCDU\n";
	print "\t                     . if not set, =CADU is assumed by default\n";
	print "\t-out=yyy             output frames type\n";
	print "\t                     . yyy = CADU | VCDU | NONE\n";
	print "\t                     . if not set, output type will be same as input type\n";
	print "\t                     . =CADU cannot be set with -in=VCDU (no encoding function)\n";
	print "\t                     . =NONE no output file will be produced, only statistics\n";
	print "\t-in_scrambled        input frames are scrambled\n";
	print "\t                     . cannot be set with -in=VCDU (only CADUs are scrambled)\n";
	print "\t-out_scrambled       output frames will be scrambled\n";
	print "\t                     . if not set and -in_scrambled is, output will be unscrambled\n";
	print "\t                     . if set and -in_scrambled is not, output will be scrambled\n";
	print "\t                     . cannot be set with -out=VCDU (only CADUs are scrambled)\n";
	print "\t-in_annotated=zzz    input frames are annotated\n";
	print "\t                     . zzz = HRDFEP | HDR | DLR | nnn (number of bytes)\n";
	print "\t-out_annotated       output frames will remain annotated\n";
	print "\t                     . if not set and -in_annotated is, annotations will be removed\n";
	print "\t                     . cannot be set if -in_annotated is not (no annotation function)\n";
	print "\t                     . cannot be set if -in=CADU and -out=VCDU (different types)\n";
	print "\t-offset=ttt          skip bytes at beginning of file\n";
	print "\t                     . ttt = n | AUTO\n";
	print "\t                     . =n skip n bytes\n";
	print "\t                     . =AUTO search for first sync marker\n";
	print "\t                     . =AUTO cannot be set with -in=VCDU (no sync marker)\n";
	print "\t-sat=sss             satellite => auto-configures -header, -frame_size, -rs_size, -rs_type\n";
	print "\t                     . sss = S1 | S2 | S3 | S5P | S6 | AE | EC | BM\n";
	print "\t                     . =S1  configures -header=v2, -frame_size=1912, rs_size=128, rs_type=223\n";
	print "\t                     . =S2  configures -header=v2, -frame_size=1912, rs_size=128, rs_type=223\n";
	print "\t                     . =S3  configures -header=v2, -frame_size=1912, rs_size=128, rs_type=223\n";
	print "\t                     . =S5P configures -header=v2, -frame_size=1115, rs_size=160, rs_type=223\n";
	print "\t                     . =S6  configures -header=v2, -frame_size=1115, rs_size=160, rs_type=223\n";
	print "\t                     . =AE  configures -header=v1, -frame_size=1115, rs_size=160, rs_type=223\n";
	print "\t                     . =EC  configures -header=v1, -frame_size=1115, rs_size=160, rs_type=223\n";
	print "\t                     . =BM  configures -header=v2, -frame_size=1912, rs_size=128, rs_type=239\n";
	print "\t-header=vv           CCSDS frame header version\n";
	print "\t                     . vv = v1 | v2\n";
	print "\t                     . is ignored if -sat is set\n";
	print "\t                     . if neither is set, =v2 is assumed by default\n";
	print "\t                     . =v1 uses 10-bits SCID, 3-bits VCID (8 channels)\n";
	print "\t                     . =v2 uses 8-bits SCID, 6-bits VCID (64 channels)\n";
	print "\t-frame_size=ffff     frame (VCDU) size in bytes\n";
	print "\t                     . is ignored if -sat is set\n";
	print "\t                     . if neither is set, =1912 is assumed by default\n";
	print "\t-rs_size=rrr         Reed-Solomon block size in bytes\n";
	print "\t                     . is ignored if -sat is set\n";
	print "\t                     . if neither is set, =128 is assumed by default\n";
	print "\t-rs_type=yyy         Reed-Solomon type, either (255,223) or (255,239)\n";
	print "\t                     . yyy = 223 | 239\n";
	print "\t                     . is ignored if -sat is set\n";
	print "\t                     . if neither is set, =223 is assumed by default\n";
	print "\t-no_idle             discard idle frames (works if scrambled or un-scrambled)\n";
	print "\t-per_VCID            split output to 1 file per VCID\n";
	print "\t-keep_bad_sync       do not abort on bad sync marker, keep storing frames\n";
	print "\t-log                 if sync errors, report in log file with VCID and counter\n";
	print "\t-log_VCID            report in log file the sequence of VCID and counter for each frame\n";
	print "\t-flip_sync           correct HDR LVDS bug: LSB storing of the sync marker\n\n";
	
	print "EXAMPLES (with trivial file names of course):\n\n";
	
	print "\tCADU-convert.pl frames_file\n";
	print "\t. no effect (all switches off), default frames with header v2, 1912 bytes and 128 RS bytes\n\n";
	
	print "\tCADU-convert.pl -sat=EC frames_file\n";
	print "\t. processes EarthCare CADUs, frame parameters are auto-configured\n\n";
	
	print "\tCADU-convert.pl -header=V2 -frame_size=1115 -rs_size=160 frames_file\n";
	print "\t. processes CADUs, frames with header v2, 1115 bytes and 160 RS bytes\n\n";
	
	print "\tCADU-convert.pl -out_scrambled CADUs_file\n";
	print "\t. scrambles CADUs to file CADUs_file.scrambled\n\n";
	
	print "\tCADU-convert.pl -in_scrambled scrambled_CADUs_file\n";
	print "\t. unscrambles CADUs to file scrambled_CADUs_file.un-scrambled\n\n";
	
	print "\tCADU-convert.pl -in_scrambled -in_annotated=HDR -no_idle HDR_scrambled_CADUs_file\n";
	print "\t. unscrambles, removes HDR annotations, discards idle frames\n";
	print "\t. and stores to file HDR_scrambled_CADUs_file.un-scrambled.not-annotated.no_idle\n\n";
	
	print "\tCADU-convert.pl -in_scrambled -out_scrambled -no_idle scrambled_CADUs_file\n";
	print "\t. removes scrambled idle frames to file scrambled_CADUs_file.no-idle\n\n";
	
	print "\tCADU-convert.pl -out_scrambled -in_annotated=HDR -out_annotated HDR_CADUs_file\n";
	print "\t. scrambles, keeps HDR annotations to file HDR_CADUs_file.scrambled\n\n";
	
	print "\tCADU-convert.pl -in=VCDU -in_annotated=HRDFEP HRDFEP_VCDUs_file\n";
	print "\t. removes VCDU HDR annotations to file HRDFEP_VCDUs_file.not-annotated\n\n";
	
	print "\tCADU-convert.pl -in=CADU -out=VCDU -in_scrambled -in_annotated=HDR HDR_CADUs_file\n";
	print "\t. unscrambles CADUs, removes HDR annotations and converts to VCDUs\n";
	print "\t. and stores to file HDR_CADUs_file.VCDU.un-scrambled.not-annotated\n\n";
	
	print "\tCADU-convert.pl -out=NONE -in_annotated=HDR -in_scrambled HDR_scrambled_CADUs_file\n";
	print "\t. unscrambles CADUs, removes HDR annotations\n";
	print "\t. no output file is generated, used only for listing VCID statistics\n";

	die "\nTry again...\n\n";
}

#-------------------------------------------------------------------------------
# initialize options using command-line switches

print "----- options -----\n";

# derived options - initialized here, properly set in the logic statements below

$f_ext = "";		# outpout file extension
$output = 1;		# 0 = no output file, only statistics
$stats = 1;			# VCID statistics; always on (code is ready for an option, if switch is added)

$xor = 0;			# 1 = scramble / unscramble (same operation)
$B_ann = 0;			# annotation size in Bytes
$no_ann = 0;		# 1 = discard annotations
$only_VCDU = 0;		# 1 = output only VCDU part
$search_sync = 0;	# 1 = look for first sync marker (file with leading bytes)
$sync_found = 0;	# 1 = first sync marker has been found
$offset_read = 0;	# counter for leading bytes skipped
$trailer_count = 0;	# trailing bytes found (i.e. incomplete CADU)
# $no_idle = ;		# directly set by the command-line switch (see below)

# ----- input and output data types -----

if ($in eq "CADU")
{
	print "input file assumed to contain CADUs\n";
}
elsif ($in eq "VCDU")
{
	print "input file assumed to to contain VCDUs\n";
}
elsif ($in eq "")		# CADUs assumed by default
{
	print "input file assumed to contain CADUs\n";
	$in = "CADU";
}
else
{
	die "\ninput file type '$in' unknown, aborting.\n\n";
}

if ($out eq "CADU")
{
	print "output file will contain CADUs\n";

	if ($in eq "VCDU") { die "\nVCDU input cannot be converted to CADU output, aborting.\n\n"; }
}
elsif ($out eq "VCDU")
{
	print "output file will contain VCDUs\n";

	if ($in eq "CADU")		# converting CADUs to VCDUs
	{
		$f_ext .= ".VCDU";
		$only_VCDU = 1;
		
		print "(if input data are scrambled <I don't know for sure>, make sure to un-scramble)\n";
	}
}
elsif ($out eq "")			# no conversion assumed by default
{
	print "output file will contain same data type as input\n";
	$out = $in;
}
elsif ($out eq "NONE")		# no output file
{
	print "output file is NOT requested\n";
	$output = 0;
}
else
{
	die "\noutput file type '$out' unknown, aborting.\n\n";
}

# ----- scrambling -----

if ($in_scrambled)
{
	print "input file assumed to be scrambled\n";
	
	if ($in eq "VCDU")
	{
		die "\ninput VCDU cannot be scrambled, aborting.\n\n";
	}

	if ($output)
	{
		if ($out_scrambled) 
		{
			print "output file will remain scrambled\n";

			if ($out eq "VCDU")
			{
				die "\noutput VCDU cannot be scrambled, aborting.\n\n";
			}
		}
		else
		{
			print "output file will be un-scrambled\n";
			$f_ext .= ".un-scrambled";
			$xor = 1;
		}
	}
}
else	# $in_scrambled = 0
{
	print "input file assumed to be un-scrambled\n";
	
	if ($output)
	{
		if ($out_scrambled) 
		{
			print "output file will be scrambled\n";
			$f_ext .= ".scrambled";
			$xor = 1;
		
			if ($out eq "VCDU")
			{
				die "\noutput VCDU cannot be scrambled, aborting.\n\n";
			}
		}
		else
		{
			print "output file will remain un-scrambled\n";
		}
	}
}

# ----- annotations -----

if ($in_annotated eq "HDR")
{	
	if ($in eq "CADU")		{ $B_ann = $B_ann_HDR_CADU; }
	elsif ($in eq "VCDU")	{ $B_ann = $B_ann_HDR_VCDU; }
	else { die "\ninput file type '$in' unknown, aborting.\n\n"; }

	print "input file assumed to have $in_annotated annotations ($B_ann bytes)\n";
}
elsif ($in_annotated eq "HRDFEP")
{
	if ($in eq "CADU")		{ $B_ann = $B_ann_HRDFEP_CADU; }
	elsif ($in eq "VCDU")	{ $B_ann = $B_ann_HRDFEP_VCDU; }
	else { die "\ninput file type '$in' unknown, aborting.\n\n"; }

	print "input file assumed to have $in_annotated annotations ($B_ann bytes)\n";
}
elsif ($in_annotated eq "DLR")
{
	if ($in eq "CADU")		{ $B_ann = $B_ann_DLR_CADU; }
	elsif ($in eq "VCDU")	{ $B_ann = $B_ann_DLR_VCDU; }
	else { die "\ninput file type unknown, aborting.\n\n"; }

	print "input file assumed to have $in_annotated annotations ($B_ann bytes)\n";
}
elsif ($in_annotated eq $in_annotated+0)	# number of bytes
{
	$B_ann = $in_annotated;
	
	print "input file assumed to have unknown annotations ($B_ann bytes)\n";
}
elsif ($in_annotated eq "")
{
	print "input file assumed to have no annotations\n";
}
else
{
	die "\ninput file annotations definition '$in_annotated' unknown, aborting.\n\n";
}

if ($output)
{
	if ($out_annotated)
	{
		print "output file will have same annotations as input\n";
	
		if ($in_annotated eq "") { die "\ninput file has no annotations, aborting.\n\n"; }
	
		if (($in eq "CADU") && ($out eq "VCDU")) { die "\nCADUs and VCDUs have incompatible 	annotations, aborting.\n\n"; }
	}
	else
	{
		if ($in_annotated)
		{
			print "output file will discard annotations\n";
			$f_ext .= ".no-annotation";
			$no_ann = 1;
		}
		else
		{
			print "output file will remain without annotations\n";
		}
	}
}

# ----- CCSDS parameters -----

if ($sat eq "S1")
{
	$header     = "v2";
	$frame_size = 1912;
	$rs_size    = 128;
	$rs_type    = 223;
	print "input file assumed to come from S1 (Sentinel-1)\n";
}
elsif ($sat eq "S2")
{
	$header     = "v2";
	$frame_size = 1912;
	$rs_size    = 128;
	$rs_type    = 223;
	print "input file assumed to come from S2 (Sentinel-2)\n";
}
elsif ($sat eq "S3")
{
	$header     = "v2";
	$frame_size = 1912;
	$rs_size    = 128;
	$rs_type    = 223;
	print "input file assumed to come from S3 (Sentinel-3)\n";
}
elsif ($sat eq "EC")
{
	$header     = "v1";
	$frame_size = 1115;
	$rs_size    = 160;
	$rs_type    = 223;
	print "input file assumed to come from EC (EarthCare)\n";
}
elsif ($sat eq "S5P")
{
	$header     = "v2";
	$frame_size = 1115;
	$rs_size    = 160;
	$rs_type    = 223;
	print "input file assumed to come from S5P (Sentinel-5P)\n";
}
elsif ($sat eq "S6")
{
	$header     = "v2";
	$frame_size = 1115;
	$rs_size    = 160;
	$rs_type    = 223;
	print "input file assumed to come from S6 (Sentinel-6)\n";
}
elsif ($sat eq "AE")
{
	$header     = "v1";
	$frame_size = 1115;
	$rs_size    = 160;
	$rs_type    = 223;
	print "input file assumed to come from AE (Aeolus)\n";
}
elsif ($sat eq "BM")
{
	$header     = "v2";
	$frame_size = 1912;
	$rs_size    = 128;
	$rs_type    = 999;
	print "input file assumed to come from BM (Biomass)\n";
}
elsif ($sat eq "")			# all this is ignored if -sat is defined
{
	if ($header eq "v1")
	{
	}
	elsif ($header eq "v2")
	{
	}
	elsif ($header eq "")
	{
		$header = "v2";
	}
	else
	{
		die "\nheader version '$header' unknown, aborting.\n\n";
	}
	
	if ($frame_size)
	{
		$frame_size +=0;
	}
	else
	{
		$frame_size = 1912;
	}
	
	if ($rs_size)
	{
		$rs_size +=0;
	}
	else
	{
		$rs_size = 128;
	}
	
	if ($rs_type)
	{
		$rs_type +=0;
	}
	else
	{
		$rs_type = 223;
	}
	print "input file origin unspecified (default CCSDS parameters will be used)\n";
}
else
{
	die "\nsatellite '$sat' unknown, aborting.\n\n";
}


if ($header eq "v1")
{
	print "input file assumed to have v1 frame headers (10-bits SCID, 3-bits VCID i.e. 8 virtual channels)\n";
	$mask_CCSDS  = "\xc0\x00";		# CCSDS version mask in byte-string hex (2 bits, bits 0-1)
	$shift_CCSDS = 14;				# CCSDS version bit shift
	$mask_SCID   = "\x3f\xf0";		# spacecraft ID mask in byte-string hex (10 bits, bits 2-11)
	$shift_SCID  = 4;				# spacecraft ID bit shift
	$mask_VCID   = "\x00\x0e";		# idle frame mask in byte-string hex (3 bits, bits 12-14)
	$shift_VCID  = 1;				# virtual channel ID bit shift
	$idle_VCID   = "\x00\x0e";		# idle frame value in byte-string hex (7, bits 12-14)
	$maxVC = 8;
}
elsif ($header eq "v2")
{
	print "input file assumed to have v2 frame headers (8-bits SCID, 6-bits VCID i.e. 64 virtual channels)\n";
	$mask_CCSDS  = "\xc0\x00";		# CCSDS version mask in byte-string hex (2 bits, bits 0-1)
	$shift_CCSDS = 14;				# CCSDS version bit shift
	$mask_SCID   = "\x3f\xc0";		# spacecraft ID mask in byte-string hex (8 bits, bits 2-9)
	$shift_SCID  = 6;				# spacecraft ID bit shift
	$mask_VCID   = "\x00\x3f";		# idle frame mask in byte-string hex (6 bits, bits 10-15)
	$shift_VCID  = 0;				# virtual channel ID bit shift
	$idle_VCID   = "\x00\x3f";		# idle frame value in byte-string hex (63, bits 10-15)
	$maxVC = 64;
}

$B_VCDU = $frame_size;
print "input file assumed to have $B_VCDU bytes frames (VCDUs)\n";
$B_RS = $rs_size;
print "input file assumed to have $B_RS bytes RS blocks\n";

# ----- other options -----

if ($offset eq "AUTO")
{
	print "input file will be searched for 1st sync marker\n";
	$f_ext .= ".skip-offset";
	$search_sync = 1;

	if ($in eq "VCDU") { die "\nVCDUs do not have sync markers, aborting.\n\n"; }
}
else
{
	$offset +=0;
	print "input file assumed to start after $offset bytes offset\n";
	if ($offset) { $f_ext .= ".skip-offset"; }
}

if ($keep_bad_sync)
{
	if ($in eq "CADU")
	{
		print "will not abort if bad sync marker found, will search next sync marker\n";
	}
	else
	{
		print "re-sync option will be ignored (input is not CADUs)\n";
	}
}

if ($output)
{
	if ($no_idle)
	{
		print "output file will discard idle frames\n";
		$f_ext .= ".no-idle";
	
		print "(if input data are scrambled <I don't know for sure> make sure to mention it)\n";
	}
}

if ($output)
{
	if ($per_VCID)
	{
		print "1 output file per VCID will be produced\n";
		$f_ext .= ".VCIDxx";
	}
}

if ($output)
{
	if ($flip_sync)
	{
		print "output file will correct for sync marker LSB storage bug\n";
		$f_ext .= ".flip-sync";
	}
}

# if no options require any processing, stop now

if ($output)
{
	if ($f_ext eq "") 
	{
		if ($stats) 
		{
			$output = 0;
			print "\noutput file would be identical to input file";
		}
		else 
		{ 
			die "\nnothing to do, aborting.\n\n";
		}
	}
}

if (!$output)	{ print "\nno output file will be produced, only statistics listed\n"; }

#-------------------------------------------------------------------------------
# use remaining command-line arguments

print "----- files -----\n";

# open put file

$f_inFRAMEs  = $ARGV[0];
print "input file:    ".$f_inFRAMEs."\n";
open(inFRAMEs,  '<:raw', $f_inFRAMEs)  || die "\nCan't open file ".$f_inFRAMEs.", aborting.\n\n";

# open output file
# if $per_VCID, open an array of output files (empty files, if any, will be deleted at the end)

if ($output)
{
	$f_outFRAMEs = $f_inFRAMEs.$f_ext;
	
	if ($per_VCID)
	{
		print "output files:  ".$f_outFRAMEs."\n";
		print "(xx can be 00 to ".($maxVC-1).", only non-empty files will be produced)\n";
		
		for ($VC = 0; $VC < $maxVC; $VC++)
		{
			$f_VC = sprintf "%02i", $VC;
			($f_outFRAMEs[$VC] = $f_outFRAMEs) =~ s|xx|$f_VC|;
			open($outFRAMEs[$VC], '>:raw', $f_outFRAMEs[$VC]) || die "\nCan't open file ".$f_outFRAMEs[$VC].", aborting.\n\n";
		}
	}
	else
	{
		print "output file:   ".$f_outFRAMEs."\n";
		open(outFRAMEs, '>:raw', $f_outFRAMEs) || die "\nCan't open file ".$f_outFRAMEs.", aborting.\n\n";
	}
}

if ($log)
{
	$f_log = $f_inFRAMEs.".errlog.tsv";
	print "log file:   ".$f_log."\n";
	open(LOG, '>:raw', $f_log) || die "\nCan't open file ".$f_log.", aborting.\n\n";

	print LOG "offset\tbefore: VCID\tbefore: VC counter\tafter: VCID\tafter: VC counter\n";
}

if ($log_VCID)
{
	$f_log_VCID = $f_inFRAMEs.".VCIDlog.tsv";
	print "VCID log file:   ".$f_log_VCID."\n";
	open(LOG_VCID, '>:raw', $f_log_VCID) || die "\nCan't open file ".$f_log_VCID.", aborting.\n\n";

	print LOG_VCID "#frame count\tVCID\tVC counter\n";
}

#-------------------------------------------------------------------------------
# derived configuration (do not touch)

$B_payload 	= $B_VCDU + $B_RS;		# all-but-sync-marker size in Bytes
$B_CADU    	= $B_SM + $B_payload;	# CADU size in Bytes
$B_annCADU  = $B_CADU + $B_ann;		# annotated CADU size in Bytes
$B_annVCDU  = $B_VCDU + $B_ann;		# annotated VCDU size in Bytes

$b_SM		= $B_SM * 8;			# Sync Marker size in bits

$correct_SM_value = vec $correct_SM, 0, $b_SM;	# convert bits (NB: not Bytes) to decimal

#-------------------------------------------------------------------------------
# scrambling polynom values, pre-computed
#
# RS(255,223) polynom is x8+x7+x5+x3+x
# init sequence is all 1's
# see e.g. http://www.noaasis.noaa.gov/LRIT/pdf-files/5_LRIT_Mission-data.pdf
#
# RS(255,239) polynom is x15+x14+1
# init sequence is ...
#
# repeating sequence of 255 bytes
# covering 1 full CADU i.e. 2040 bytes (sync marker excluded of course)

if ($rs_type == 223)
{
	@polynom_255 = (
	0xff,0x48,0x0e,0xc0,0x9a,0x0d,0x70,0xbc,0x8e,0x2c,	#   10
	0x93,0xad,0xa7,0xb7,0x46,0xce,0x5a,0x97,0x7d,0xcc,	#   20
	0x32,0xa2,0xbf,0x3e,0x0a,0x10,0xf1,0x88,0x94,0xcd,	#   30
	0xea,0xb1,0xfe,0x90,0x1d,0x81,0x34,0x1a,0xe1,0x79,	#   40
	0x1c,0x59,0x27,0x5b,0x4f,0x6e,0x8d,0x9c,0xb5,0x2e,	#   50
	0xfb,0x98,0x65,0x45,0x7e,0x7c,0x14,0x21,0xe3,0x11,	#   60
	0x29,0x9b,0xd5,0x63,0xfd,0x20,0x3b,0x02,0x68,0x35,	#   70
	0xc2,0xf2,0x38,0xb2,0x4e,0xb6,0x9e,0xdd,0x1b,0x39,	#   80
	0x6a,0x5d,0xf7,0x30,0xca,0x8a,0xfc,0xf8,0x28,0x43,	#   90
	0xc6,0x22,0x53,0x37,0xaa,0xc7,0xfa,0x40,0x76,0x04,	#  100
	0xd0,0x6b,0x85,0xe4,0x71,0x64,0x9d,0x6d,0x3d,0xba,	#  110
	0x36,0x72,0xd4,0xbb,0xee,0x61,0x95,0x15,0xf9,0xf0,	#  120
	0x50,0x87,0x8c,0x44,0xa6,0x6f,0x55,0x8f,0xf4,0x80,	#  130
	0xec,0x09,0xa0,0xd7,0x0b,0xc8,0xe2,0xc9,0x3a,0xda,	#  140
	0x7b,0x74,0x6c,0xe5,0xa9,0x77,0xdc,0xc3,0x2a,0x2b,	#  150
	0xf3,0xe0,0xa1,0x0f,0x18,0x89,0x4c,0xde,0xab,0x1f,	#  160
	0xe9,0x01,0xd8,0x13,0x41,0xae,0x17,0x91,0xc5,0x92,	#  170
	0x75,0xb4,0xf6,0xe8,0xd9,0xcb,0x52,0xef,0xb9,0x86,	#  180
	0x54,0x57,0xe7,0xc1,0x42,0x1e,0x31,0x12,0x99,0xbd,	#  190
	0x56,0x3f,0xd2,0x03,0xb0,0x26,0x83,0x5c,0x2f,0x23,	#  200
	0x8b,0x24,0xeb,0x69,0xed,0xd1,0xb3,0x96,0xa5,0xdf,	#  210
	0x73,0x0c,0xa8,0xaf,0xcf,0x82,0x84,0x3c,0x62,0x25,	#  220
	0x33,0x7a,0xac,0x7f,0xa4,0x07,0x60,0x4d,0x06,0xb8,	#  230
	0x5e,0x47,0x16,0x49,0xd6,0xd3,0xdb,0xa3,0x67,0x2d,	#  240
	0x4b,0xbe,0xe6,0x19,0x51,0x5f,0x9f,0x05,0x08,0x78,	#  250
	0xc4,0x4a,0x66,0xf5,0x58							#  255
	);
	push @polynom, @polynom_255 for (1..8);				# fill up to 2040 bytes

}
else	# BioMass (only, so far)
{
	@polynom_2040 = (
	0x03,0xF6,0x08,0x34,0x30,0xB8,0xA3,0x93,0xC9,0x68,	#   10
	0xB7,0x73,0xB3,0x29,0xAA,0xF5,0xFE,0x3C,0x04,0x88,	#   20
	0x1B,0x30,0x5A,0xA1,0xDF,0xC4,0xC0,0x9A,0x83,0x5F,	#   30
	0x0B,0xC2,0x38,0x8C,0x93,0x2B,0x6A,0xFB,0x7E,0x1B,	#   40
	0x04,0x5A,0x19,0xDC,0x54,0xC9,0xFA,0xB4,0x1F,0xB8,	#   50
	0x41,0x91,0x85,0x65,0x1F,0x5E,0x43,0xC5,0x88,0x9D,	#   60
	0x33,0x4E,0xAB,0xA7,0xF9,0xD0,0x14,0xE0,0x7A,0x41,	#   70
	0x1D,0x86,0x4D,0x15,0xAE,0x7D,0xE5,0x0C,0x5E,0x29,	#   80
	0xC4,0xF4,0x9A,0x3B,0x5C,0x9B,0xCB,0x58,0xBB,0xD3,	#   90
	0x98,0xE9,0x52,0x77,0xED,0x30,0x6E,0xA1,0x67,0xC7,	#  100
	0x50,0x93,0xE3,0x68,0x4B,0x71,0xBB,0x25,0x9A,0xDD,	#  110
	0x5E,0xCF,0xC6,0xA0,0x97,0xC3,0x70,0x8B,0x23,0x3A,	#  120
	0xCA,0x9E,0xBF,0x47,0x83,0x91,0x09,0x66,0x37,0x54,	#  130
	0xB3,0xFB,0xA8,0x19,0xF0,0x54,0x21,0xF8,0xC4,0x12,	#  140
	0x98,0x6F,0x51,0x63,0xE7,0x48,0x53,0xB1,0xE9,0xA4,	#  150
	0x75,0xD9,0x3C,0xD6,0x8A,0xF7,0x3E,0x32,0x84,0xAF,	#  160
	0x1B,0xE2,0x58,0x4D,0xD1,0xAC,0xE5,0xEA,0x5C,0x7D,	#  170
	0xC9,0x0C,0xB6,0x2B,0xB4,0xF9,0xBA,0x15,0x9C,0x7D,	#  180
	0x49,0x0F,0xB6,0x21,0xB4,0xC5,0xBA,0x9D,0x9F,0x4D,	#  190
	0x43,0xAF,0x89,0xE1,0x34,0x46,0xB9,0x97,0x95,0x71,	#  200
	0x7F,0x27,0x02,0xD2,0x0E,0xEC,0x26,0x68,0xD5,0x72,	#  210
	0xFF,0x2E,0x02,0xE4,0x0E,0x58,0x25,0xD0,0xDC,0xE2,	#  220
	0xCA,0x4E,0xBD,0xA7,0x8D,0xD1,0x2C,0xE6,0xEA,0x56,	#  230
	0x7D,0xF5,0x0C,0x3E,0x28,0x84,0xF3,0x1A,0x2A,0x5C,	#  240
	0xFD,0xCA,0x0C,0xBC,0x2B,0x88,0xF9,0x32,0x16,0xAC,	#  250
	0x77,0xE9,0x30,0x76,0xA1,0x37,0xC6,0xB0,0x97,0xA3,	#  260
	0x71,0xCB,0x24,0xBA,0xDB,0x9E,0xD9,0x46,0xD7,0x96,	#  270
	0xF1,0x76,0x27,0x34,0xD2,0xBA,0xEF,0x9E,0x61,0x45,	#  280
	0x47,0x9F,0x91,0x41,0x67,0x87,0x51,0x13,0xE6,0x68,	#  290
	0x55,0x71,0xFF,0x24,0x02,0xD8,0x0E,0xD0,0x26,0xE0,	#  300
	0xD6,0x42,0xF5,0x8E,0x3D,0x24,0x8E,0xDB,0x26,0xDA,	#  310
	0xD6,0xDE,0xF6,0xC6,0x36,0x94,0xB7,0x7B,0xB3,0x19,	#  320
	0xAA,0x55,0xFD,0xFC,0x0C,0x08,0x28,0x30,0xF0,0xA2,	#  330
	0x23,0xCC,0xC8,0xAA,0xB3,0xFF,0xA8,0x01,0xF0,0x04,	#  340
	0x20,0x18,0xC0,0x52,0x81,0xEF,0x04,0x62,0x19,0x4C,	#  350
	0x57,0xA9,0xF1,0xF4,0x24,0x38,0xD8,0x92,0xD3,0x6E,	#  360
	0xEB,0x66,0x7B,0x55,0x1B,0xFE,0x58,0x05,0xD0,0x1C,	#  370
	0xE0,0x4A,0x41,0xBD,0x85,0x8D,0x1D,0x2E,0x4E,0xE5,	#  380
	0xA6,0x5D,0xD5,0xCC,0xFC,0xAA,0x0B,0xFC,0x38,0x08,	#  390
	0x90,0x33,0x60,0xAB,0x43,0xFB,0x88,0x19,0x30,0x56,	#  400
	0xA1,0xF7,0xC4,0x30,0x98,0xA3,0x53,0xCB,0xE8,0xB8,	#  410
	0x73,0x91,0x29,0x66,0xF7,0x56,0x33,0xF4,0xA8,0x3B,	#  420
	0xF0,0x98,0x23,0x50,0xCB,0xE2,0xB8,0x4F,0x91,0xA1,	#  430
	0x65,0xC7,0x5C,0x93,0xCB,0x68,0xBB,0x73,0x9B,0x29,	#  440
	0x5A,0xF7,0xDE,0x30,0xC4,0xA2,0x9B,0xCF,0x58,0xA3,	#  450
	0xD3,0xC8,0xE8,0xB2,0x73,0xAD,0x29,0xEE,0xF4,0x66,	#  460
	0x39,0x54,0x97,0xFB,0x70,0x1B,0x20,0x5A,0xC1,0xDE,	#  470
	0x84,0xC7,0x1A,0x92,0x5F,0x6D,0xC3,0x6C,0x8B,0x6B,	#  480
	0x3B,0x7A,0x9B,0x1F,0x5A,0x43,0xDD,0x88,0xCD,0x32,	#  490
	0xAE,0xAF,0xE7,0xE0,0x50,0x41,0xE1,0x84,0x45,0x19,	#  500
	0x9E,0x55,0x45,0xFF,0x9C,0x01,0x48,0x07,0xB0,0x11,	#  510
	0xA0,0x65,0xC1,0x5C,0x87,0xCB,0x10,0xBA,0x63,0x9D,	#  520
	0x49,0x4F,0xB7,0xA1,0xB1,0xC5,0xA4,0x9D,0xDB,0x4C,	#  530
	0xDB,0xAA,0xD9,0xFE,0xD4,0x06,0xF8,0x16,0x10,0x74,	#  540
	0x61,0x39,0x46,0x97,0x97,0x71,0x73,0x27,0x2A,0xD2,	#  550
	0xFE,0xEE,0x06,0x64,0x15,0x58,0x7F,0xD1,0x00,0xE6,	#  560
	0x02,0x54,0x0D,0xF8,0x2C,0x10,0xE8,0x62,0x71,0x4D,	#  570
	0x27,0xAE,0xD1,0xE6,0xE4,0x56,0x59,0xF5,0xD4,0x3C,	#  580
	0xF8,0x8A,0x13,0x3C,0x6A,0x89,0x7F,0x37,0x02,0xB2,	#  590
	0x0F,0xAC,0x21,0xE8,0xC4,0x72,0x99,0x2F,0x56,0xE3,	#  600
	0xF6,0x48,0x35,0xB0,0xBD,0xA3,0x8D,0xC9,0x2C,0xB6,	#  610
	0xEB,0xB6,0x79,0xB5,0x15,0xBE,0x7D,0x85,0x0D,0x1E,	#  620
	0x2E,0x44,0xE5,0x9A,0x5D,0x5D,0xCF,0xCC,0xA0,0xAB,	#  630
	0xC3,0xF8,0x88,0x13,0x30,0x6A,0xA1,0x7F,0xC7,0x00,	#  640
	0x92,0x03,0x6C,0x0B,0x68,0x3B,0x70,0x9B,0x23,0x5A,	#  650
	0xCB,0xDE,0xB8,0xC7,0x92,0x91,0x6F,0x67,0x63,0x53,	#  660
	0x4B,0xEB,0xB8,0x79,0x91,0x15,0x66,0x7F,0x55,0x03,	#  670
	0xFE,0x08,0x04,0x30,0x18,0xA0,0x53,0xC1,0xE8,0x84,	#  680
	0x73,0x19,0x2A,0x56,0xFD,0xF6,0x0C,0x34,0x28,0xB8,	#  690
	0xF3,0x92,0x29,0x6C,0xF7,0x6A,0x33,0x7C,0xAB,0x0B,	#  700
	0xFA,0x38,0x1C,0x90,0x4B,0x61,0xBB,0x45,0x9B,0x9D,	#  710
	0x59,0x4F,0xD7,0xA0,0xF1,0xC2,0x24,0x8C,0xDB,0x2A,	#  720
	0xDA,0xFE,0xDE,0x06,0xC4,0x16,0x98,0x77,0x51,0x33,	#  730
	0xE6,0xA8,0x57,0xF1,0xF0,0x24,0x20,0xD8,0xC2,0xD2,	#  740
	0x8E,0xEF,0x26,0x62,0xD5,0x4E,0xFF,0xA6,0x01,0xD4,	#  750
	0x04,0xF8,0x1A,0x10,0x5C,0x61,0xC9,0x44,0xB7,0x9B,	#  760
	0xB1,0x59,0xA7,0xD5,0xD0,0xFC,0xE2,0x0A,0x4C,0x3D,	#  770
	0xA8,0x8D,0xF3,0x2C,0x2A,0xE8,0xFE,0x72,0x05,0x2C,	#  780
	0x1E,0xE8,0x46,0x71,0x95,0x25,0x7E,0xDF,0x06,0xC2,	#  790
	0x16,0x8C,0x77,0x29,0x32,0xF6,0xAE,0x37,0xE4,0xB0,	#  800
	0x5B,0xA1,0xD9,0xC4,0xD4,0x9A,0xFB,0x5E,0x1B,0xC4,	#  810
	0x58,0x99,0xD3,0x54,0xEB,0xFA,0x78,0x1D,0x10,0x4E,	#  820
	0x61,0xA5,0x45,0xDF,0x9C,0xC1,0x4A,0x87,0xBF,0x11,	#  830
	0x82,0x65,0x0D,0x5E,0x2F,0xC4,0xE0,0x9A,0x43,0x5D,	#  840
	0x8B,0xCD,0x38,0xAE,0x93,0xE7,0x68,0x53,0x71,0xEB,	#  850
	0x24,0x7A,0xD9,0x1E,0xD6,0x46,0xF5,0x96,0x3D,0x74,	#  860
	0x8F,0x3B,0x22,0x9A,0xCF,0x5E,0xA3,0xC7,0xC8,0x90,	#  870
	0xB3,0x63,0xAB,0x49,0xFB,0xB4,0x19,0xB8,0x55,0x91,	#  880
	0xFD,0x64,0x0F,0x58,0x23,0xD0,0xC8,0xE2,0xB2,0x4F,	#  890
	0xAD,0xA1,0xED,0xC4,0x6C,0x99,0x6B,0x57,0x7B,0xF3,	#  900
	0x18,0x2A,0x50,0xFD,0xE2,0x0C,0x4C,0x29,0xA8,0xF5,	#  910
	0xF2,0x3C,0x2C,0x88,0xEB,0x32,0x7A,0xAD,0x1F,0xEE,	#  920
	0x40,0x65,0x81,0x5D,0x07,0xCE,0x10,0xA4,0x63,0xD9,	#  930
	0x48,0xD7,0xB2,0xF1,0xAE,0x25,0xE4,0xDC,0x5A,0xC9,	#  940
	0xDE,0xB4,0xC7,0xBA,0x91,0x9F,0x65,0x43,0x5F,0x8B,	#  950
	0xC1,0x38,0x86,0x93,0x17,0x6A,0x73,0x7D,0x2B,0x0E,	#  960
	0xFA,0x26,0x1C,0xD4,0x4A,0xF9,0xBE,0x15,0x84,0x7D,	#  970
	0x19,0x0E,0x56,0x25,0xF4,0xDC,0x3A,0xC8,0x9E,0xB3,	#  980
	0x47,0xAB,0x91,0xF9,0x64,0x17,0x58,0x73,0xD1,0x28,	#  990
	0xE6,0xF2,0x56,0x2D,0xF4,0xEC,0x3A,0x68,0x9D,0x73,	# 1000
	0x4F,0x2B,0xA2,0xF9,0xCE,0x14,0xA4,0x7B,0xD9,0x18,	# 1010
	0xD6,0x52,0xF5,0xEE,0x3C,0x64,0x89,0x5B,0x37,0xDA,	# 1020
	0xB0,0xDF,0xA2,0xC1,0xCE,0x84,0xA7,0x1B,0xD2,0x58,	# 1030
	0xED,0xD2,0x6C,0xED,0x6A,0x6F,0x7D,0x63,0x0F,0x4A,	# 1040
	0x23,0xBC,0xC9,0x8A,0xB5,0x3F,0xBE,0x81,0x87,0x05,	# 1050
	0x12,0x1E,0x6C,0x45,0x69,0x9F,0x75,0x43,0x3F,0x8A,	# 1060
	0x81,0x3F,0x06,0x82,0x17,0x0C,0x72,0x29,0x2C,0xF6,	# 1070
	0xEA,0x36,0x7C,0xB5,0x0B,0xBE,0x39,0x84,0x95,0x1B,	# 1080
	0x7E,0x5B,0x05,0xDA,0x1C,0xDC,0x4A,0xC9,0xBE,0xB5,	# 1090
	0x87,0xBD,0x11,0x8E,0x65,0x25,0x5E,0xDF,0xC6,0xC0,	# 1100
	0x96,0x83,0x77,0x0B,0x32,0x3A,0xAC,0x9F,0xEB,0x40,	# 1110
	0x7B,0x81,0x19,0x06,0x56,0x15,0xF4,0x7C,0x39,0x08,	# 1120
	0x96,0x33,0x74,0xAB,0x3B,0xFA,0x98,0x1F,0x50,0x43,	# 1130
	0xE1,0x88,0x45,0x31,0x9E,0xA5,0x47,0xDF,0x90,0xC1,	# 1140
	0x62,0x87,0x4F,0x13,0xA2,0x69,0xCD,0x74,0xAF,0x3B,	# 1150
	0xE2,0x98,0x4F,0x51,0xA3,0xE5,0xC8,0x5C,0xB1,0xCB,	# 1160
	0xA4,0xB9,0xDB,0x94,0xD9,0x7A,0xD7,0x1E,0xF2,0x46,	# 1170
	0x2D,0x94,0xED,0x7A,0x6F,0x1D,0x62,0x4F,0x4D,0xA3,	# 1180
	0xAD,0xC9,0xEC,0xB4,0x6B,0xB9,0x79,0x97,0x15,0x72,	# 1190
	0x7F,0x2D,0x02,0xEE,0x0E,0x64,0x25,0x58,0xDF,0xD2,	# 1200
	0xC0,0xEE,0x82,0x67,0x0D,0x52,0x2F,0xEC,0xE0,0x6A,	# 1210
	0x41,0x7D,0x87,0x0D,0x12,0x2E,0x6C,0xE5,0x6A,0x5F,	# 1220
	0x7D,0xC3,0x0C,0x8A,0x2B,0x3C,0xFA,0x8A,0x1F,0x3C,	# 1230
	0x42,0x89,0x8F,0x35,0x22,0xBE,0xCF,0x86,0xA1,0x17,	# 1240
	0xC6,0x70,0x95,0x23,0x7E,0xCB,0x06,0xBA,0x17,0x9C,	# 1250
	0x71,0x49,0x27,0xB6,0xD1,0xB6,0xE5,0xB6,0x5D,0xB5,	# 1260
	0xCD,0xBC,0xAD,0x8B,0xED,0x38,0x6E,0x91,0x67,0x67,	# 1270
	0x53,0x53,0xEB,0xE8,0x78,0x71,0x11,0x26,0x66,0xD5,	# 1280
	0x56,0xFF,0xF6,0x00,0x34,0x00,0xB8,0x03,0x90,0x09,	# 1290
	0x60,0x37,0x40,0xB3,0x83,0xA9,0x09,0xF6,0x34,0x34,	# 1300
	0xB8,0xBB,0x93,0x99,0x69,0x57,0x77,0xF3,0x30,0x2A,	# 1310
	0xA0,0xFF,0xC2,0x00,0x8C,0x03,0x28,0x0A,0xF0,0x3E,	# 1320
	0x20,0x84,0xC3,0x1A,0x8A,0x5F,0x3D,0xC2,0x8C,0x8F,	# 1330
	0x2B,0x22,0xFA,0xCE,0x1E,0xA4,0x47,0xD9,0x90,0xD5,	# 1340
	0x62,0xFF,0x4E,0x03,0xA4,0x09,0xD8,0x34,0xD0,0xBA,	# 1350
	0xE3,0x9E,0x49,0x45,0xB7,0x9D,0xB1,0x4D,0xA7,0xAD,	# 1360
	0xD1,0xEC,0xE4,0x6A,0x59,0x7D,0xD7,0x0C,0xF2,0x2A,	# 1370
	0x2C,0xFC,0xEA,0x0A,0x7C,0x3D,0x08,0x8E,0x33,0x24,	# 1380
	0xAA,0xDB,0xFE,0xD8,0x06,0xD0,0x16,0xE0,0x76,0x41,	# 1390
	0x35,0x86,0xBD,0x17,0x8E,0x71,0x25,0x26,0xDE,0xD6,	# 1400
	0xC6,0xF6,0x96,0x37,0x74,0xB3,0x3B,0xAA,0x99,0xFF,	# 1410
	0x54,0x03,0xF8,0x08,0x10,0x30,0x60,0xA1,0x43,0xC7,	# 1420
	0x88,0x91,0x33,0x66,0xAB,0x57,0xFB,0xF0,0x18,0x20,	# 1430
	0x50,0xC1,0xE2,0x84,0x4F,0x19,0xA2,0x55,0xCD,0xFC,	# 1440
	0xAC,0x0B,0xE8,0x38,0x70,0x91,0x23,0x66,0xCB,0x56,	# 1450
	0xBB,0xF7,0x98,0x31,0x50,0xA7,0xE3,0xD0,0x48,0xE1,	# 1460
	0xB2,0x45,0xAD,0x9D,0xED,0x4C,0x6F,0xA9,0x61,0xF7,	# 1470
	0x44,0x33,0x98,0xA9,0x53,0xF7,0xE8,0x30,0x70,0xA1,	# 1480
	0x23,0xC6,0xC8,0x96,0xB3,0x77,0xAB,0x31,0xFA,0xA4,	# 1490
	0x1F,0xD8,0x40,0xD1,0x82,0xE5,0x0E,0x5E,0x25,0xC4,	# 1500
	0xDC,0x9A,0xCB,0x5E,0xBB,0xC7,0x98,0x91,0x53,0x67,	# 1510
	0xEB,0x50,0x7B,0xE1,0x18,0x46,0x51,0x95,0xE5,0x7C,	# 1520
	0x5F,0x09,0xC2,0x34,0x8C,0xBB,0x2B,0x9A,0xF9,0x5E,	# 1530
	0x17,0xC4,0x70,0x99,0x23,0x56,0xCB,0xF6,0xB8,0x37,	# 1540
	0x90,0xB1,0x63,0xA7,0x49,0xD3,0xB4,0xE9,0xBA,0x75,	# 1550
	0x9D,0x3D,0x4E,0x8F,0xA7,0x21,0xD2,0xC4,0xEE,0x9A,	# 1560
	0x67,0x5D,0x53,0xCF,0xE8,0xA0,0x73,0xC1,0x28,0x86,	# 1570
	0xF3,0x16,0x2A,0x74,0xFD,0x3A,0x0E,0x9C,0x27,0x48,	# 1580
	0xD3,0xB2,0xE9,0xAE,0x75,0xE5,0x3C,0x5E,0x89,0xC7,	# 1590
	0x34,0x92,0xBB,0x6F,0x9B,0x61,0x5B,0x47,0xDB,0x90,	# 1600
	0xD9,0x62,0xD7,0x4E,0xF3,0xA6,0x29,0xD4,0xF4,0xFA,	# 1610
	0x3A,0x1C,0x9C,0x4B,0x49,0xBB,0xB5,0x99,0xBD,0x55,	# 1620
	0x8F,0xFD,0x20,0x0E,0xC0,0x26,0x80,0xD7,0x02,0xF2,	# 1630
	0x0E,0x2C,0x24,0xE8,0xDA,0x72,0xDD,0x2E,0xCE,0xE6,	# 1640
	0xA6,0x57,0xD5,0xF0,0xFC,0x22,0x08,0xCC,0x32,0xA8,	# 1650
	0xAF,0xF3,0xE0,0x28,0x40,0xF1,0x82,0x25,0x0C,0xDE,	# 1660
	0x2A,0xC4,0xFE,0x9A,0x07,0x5C,0x13,0xC8,0x68,0xB1,	# 1670
	0x73,0xA7,0x29,0xD2,0xF4,0xEE,0x3A,0x64,0x9D,0x5B,	# 1680
	0x4F,0xDB,0xA0,0xD9,0xC2,0xD4,0x8E,0xFB,0x26,0x1A,	# 1690
	0xD4,0x5E,0xF9,0xC6,0x14,0x94,0x7B,0x79,0x1B,0x16,	# 1700
	0x5A,0x75,0xDD,0x3C,0xCE,0x8A,0xA7,0x3F,0xD2,0x80,	# 1710
	0xEF,0x02,0x62,0x0D,0x4C,0x2F,0xA8,0xE1,0xF2,0x44,	# 1720
	0x2D,0x98,0xED,0x52,0x6F,0xED,0x60,0x6F,0x41,0x63,	# 1730
	0x87,0x49,0x13,0xB6,0x69,0xB5,0x75,0xBF,0x3D,0x82,	# 1740
	0x8D,0x0F,0x2E,0x22,0xE4,0xCE,0x5A,0xA5,0xDF,0xDC,	# 1750
	0xC0,0xCA,0x82,0xBF,0x0F,0x82,0x21,0x0C,0xC6,0x2A,	# 1760
	0x94,0xFF,0x7A,0x03,0x1C,0x0A,0x48,0x3D,0xB0,0x8D,	# 1770
	0xA3,0x2D,0xCA,0xEC,0xBE,0x6B,0x85,0x79,0x1F,0x16,	# 1780
	0x42,0x75,0x8D,0x3D,0x2E,0x8E,0xE7,0x26,0x52,0xD5,	# 1790
	0xEE,0xFC,0x66,0x09,0x54,0x37,0xF8,0xB0,0x13,0xA0,	# 1800
	0x69,0xC1,0x74,0x87,0x3B,0x12,0x9A,0x6F,0x5D,0x63,	# 1810
	0xCF,0x48,0xA3,0xB3,0xC9,0xA8,0xB5,0xF3,0xBC,0x29,	# 1820
	0x88,0xF5,0x32,0x3E,0xAC,0x87,0xEB,0x10,0x7A,0x61,	# 1830
	0x1D,0x46,0x4F,0x95,0xA1,0x7D,0xC7,0x0C,0x92,0x2B,	# 1840
	0x6C,0xFB,0x6A,0x1B,0x7C,0x5B,0x09,0xDA,0x34,0xDC,	# 1850
	0xBA,0xCB,0x9E,0xB9,0x47,0x97,0x91,0x71,0x67,0x27,	# 1860
	0x52,0xD3,0xEE,0xE8,0x66,0x71,0x55,0x27,0xFE,0xD0,	# 1870
	0x06,0xE0,0x16,0x40,0x75,0x81,0x3D,0x06,0x8E,0x17,	# 1880
	0x24,0x72,0xD9,0x2E,0xD6,0xE6,0xF6,0x56,0x35,0xF4,	# 1890
	0xBC,0x3B,0x88,0x99,0x33,0x56,0xAB,0xF7,0xF8,0x30,	# 1900
	0x10,0xA0,0x63,0xC1,0x48,0x87,0xB3,0x11,0xAA,0x65,	# 1910
	0xFD,0x5C,0x0F,0xC8,0x20,0xB0,0xC3,0xA2,0x89,0xCF,	# 1920
	0x34,0xA2,0xBB,0xCF,0x98,0xA1,0x53,0xC7,0xE8,0x90,	# 1930
	0x73,0x61,0x2B,0x46,0xFB,0x96,0x19,0x74,0x57,0x39,	# 1940
	0xF2,0x94,0x2F,0x78,0xE3,0x12,0x4A,0x6D,0xBD,0x6D,	# 1950
	0x8F,0x6D,0x23,0x6E,0xCB,0x66,0xBB,0x57,0x9B,0xF1,	# 1960
	0x58,0x27,0xD0,0xD0,0xE2,0xE2,0x4E,0x4D,0xA5,0xAD,	# 1970
	0xDD,0xEC,0xCC,0x6A,0xA9,0x7F,0xF7,0x00,0x32,0x00,	# 1980
	0xAC,0x03,0xE8,0x08,0x70,0x31,0x20,0xA6,0xC3,0xD6,	# 1990
	0x88,0xF7,0x32,0x32,0xAC,0xAF,0xEB,0xE0,0x78,0x41,	# 2000
	0x11,0x86,0x65,0x15,0x5E,0x7F,0xC5,0x00,0x9E,0x03,	# 2010
	0x44,0x0B,0x98,0x39,0x50,0x97,0xE3,0x70,0x4B,0x21,	# 2020
	0xBA,0xC5,0x9E,0x9D,0x47,0x4F,0x93,0xA1,0x69,0xC7,	# 2030
	0x74,0x93,0x3B,0x6A,0x9B,0x7F,0x5B,0x03,0xDA,0x08	# 2040
	);
	@polynom = @polynom_2040;				# fill up to 2040 bytes
}


$polynom_pack_format ="C".$B_payload;				# same size as CADU payload
$polynom_bytes		= pack($polynom_pack_format, @polynom);
$polynom_2bytes		= substr $polynom_bytes, 0, 2;		# first 2 bytes (contains VCID)
$polynom_counter	= substr $polynom_bytes, 2, 3;		# next 3 bytes (contains VC counter)

#===============================================================================

# initialisation

$n_FRAMEs = 0;		# total FRAME count
$n_niFRAMEs = 0;	# non-idle FRAME count

print "----- processing -----\n";

# skip leading bytes if any -----------------------------------------------------------

SYNC:
if ($search_sync)
{	
	$offset_read = 0;				
	$sync_found = 0;
	$bytes_read = 0;
	
	while (($bytes_read = read(inFRAMEs, $search_bytes, $B_search, 0)) > 3)	# if 3 bytes left, no sync possible
	{
		$SM_offset = index $search_bytes, $correct_SM;		# byte offset for sync marker

		if ($SM_offset == -1)		# no sync marker found in this block
		{
			$offset_read += $bytes_read - $B_SM + 1;	# count bytes skipped
			
			printf "searching for sync marker %8x... skipping $offset_read bytes", $correct_SM_value;
			print "\r";
			
			$rewind = tell(inFRAMEs) - $B_SM + 1;	# go back 3 bytes
			seek(inFRAMEs, $rewind, 0);				
		}
		else						# sync marker found in this block
		{
			$offset_read += $SM_offset;				# count bytes skipped
			
			printf "searching for sync marker %8x... found after $offset_read bytes", $correct_SM_value;
			print "\n";
			
			$rewind = tell(inFRAMEs) - $bytes_read + $SM_offset;	# go back to sync marker
			seek(inFRAMEs, $rewind, 0);

			$sync_found = 1;
			last;					# stop searching (exit loop)
		}
	}
	
	if ( ! $sync_found)
	{
		printf "\nsync marker %8x not found, aborting.\n\n", $correct_SM_value;
		die;
	}
}
else
{
	$offset_read = read(inFRAMEs, $offset_bytes, $offset, 0);
	
	$sync_found = 1;
}

print "$offset_read leading bytes skipped\n", ;

if ($log)	# log info from PREVIOUS frame
{
	if ($payload_bytes eq "")	# first sync, no previous frame yet
	{
		printf LOG "$offset_read\tn/a\tn/a\t";
	}
	else				# re-sync, previous frame data available
	{
		$counter_bytes = "\x00".substr $payload_bytes, 2, 3;	# VC counter 
		if ($out_scrambled)		# need to xor to decode VC counter
		{
			$counter_bytes = $counter_bytes ^ "\x00".$polynom_counter;
		}
		$counter_value = vec $counter_bytes, 0, 32;				# in decimal
		$VC = unpack "n", $VCID;	$VC = $VC >> $shift_VCID;

		printf LOG "$offset_read\t$VC\t$counter_value\t";
	}
}				

# read one frame at a time, process, write one frame at a time -----------------

$bytes_read = 0;
	
if ($in eq "CADU")		# input is CADUs
{
	while ($bytes_read = read(inFRAMEs, $annCADU_bytes, $B_annCADU, 0))
	{
		if ($bytes_read == $B_annCADU)
		{
			printf "CADU %8d,   $bytes_read bytes", ++$n_FRAMEs;
			
			# ----- process sync marker -----
			
			$SM_bytes = substr $annCADU_bytes, 0, $B_SM;			# sync marker only
		
			if ($flip_sync)		# need to put back sync marker in MSB
			{
				$SM_bytes = pack("B".$b_SM, unpack("b".$b_SM, $SM_bytes));
			}
	
			if ($SM_bytes eq $correct_SM)
			{
				printf ",   sync = (x) %8x", $correct_SM_value;
			}
			else		# incorrect sync marker
			{
				$SM = vec $SM_bytes, 0, $b_SM;			# convert bits (NB: not Bytes) to decimal
				printf ",   sync = (x) %8x", $SM;
				
				if ($keep_bad_sync)
				{
					# check next bytes for sync marker and rewind immediately
					$bytes_read = read(inFRAMEs, $next_SM_bytes, $B_SM, 0);
					$rewind = tell(inFRAMEs) - $B_SM;
					seek(inFRAMEs, $rewind, 0);				
					
					if ($next_SM_bytes eq $correct_SM)		# found sync marker immediately
					{
						$damaged_SM_count++;
						$sync_found = 1;
						
						if ($log)
						{
							$counter_bytes = "\x00".substr $payload_bytes, 2, 3;	# VC counter 
							if ($out_scrambled)		# need to xor to decode VC counter
							{
								$counter_bytes = $counter_bytes ^ "\x00".$polynom_counter;
							}
							$counter_value = vec $counter_bytes, 0, 32;				# in decimal
							$VC = unpack "n", $VCID;	$VC = $VC >> $shift_VCID;
						
							printf LOG "0\t$VC\t$counter_value\t";
						}
					}
					else		# no sync marker found immediately, sync again
					{
						$rewind = tell(inFRAMEs) - $B_annCADU;	# go back 1 frame
						seek(inFRAMEs, $rewind, 0);	

						print "\n";
						$search_sync = 1; goto SYNC;			# search again for sync marker
					}
				}
				else
				{
					die ",   Wrong Sync Marker, aborting.\n\n";
				}
			}
			
			# ----- retrieve CADU payload -----
		
			$payload_bytes = substr $annCADU_bytes, $B_SM, $B_payload;	# CADU without sync marker
		
			# ----- decode VCID -----
		
			$payload_2bytes = substr $payload_bytes, 0, 2;		# first 2 bytes
		
			if ($in_scrambled)		# need to xor to decode VCID
			{
				$payload_2bytes = $payload_2bytes ^ $polynom_2bytes;
			}
				
			$CCSDS =  $payload_2bytes & $mask_CCSDS ;	
			$SCID  =  $payload_2bytes & $mask_SCID ;	
			$VCID  =  $payload_2bytes & $mask_VCID ;	
			
			# ----- process scrambling of CADU payload -----
		
			if ($xor)			# need to scramble / unscramble (same operation)
			{
				$payload_bytes = $payload_bytes ^ $polynom_bytes;		# XOR to scramble/descramble
			}
		
			# ----- statistics -----
		
			$VC_count{$VCID}++;			# VCID statistics
			
			$CC = unpack "n", $CCSDS; $CC = $CC >> $shift_CCSDS; $CC++;
			$SC = unpack "n", $SCID;  $SC = $SC >> $shift_SCID;
			$VC = unpack "n", $VCID;  $VC = $VC >> $shift_VCID;
		
			if ($log_VCID)
			{
				$counter_bytes = "\x00".substr $payload_bytes, 2, 3;	# VC counter 
				if ($out_scrambled)		# need to xor to decode VC counter
				{
					$counter_bytes = $counter_bytes ^ "\x00".$polynom_counter;
				}
				$counter_value = vec $counter_bytes, 0, 32;				# in decimal

				print LOG_VCID "$n_FRAMEs\t$VC\t$counter_value\n";
			}
			# ----- log if (re-)sync found -----
		
			if (($sync_found) && ($log))	# log info from CURRENT frame
			{
				$counter_bytes = "\x00".substr $payload_bytes, 2, 3;	# VC counter 
				if ($out_scrambled)		# need to xor to decode VC counter
				{
					$counter_bytes = $counter_bytes ^ "\x00".$polynom_counter;
				}
				$counter_value = vec $counter_bytes, 0, 32;				# in decimal
				$VC = unpack "n", $VCID;	$VC = $VC >> $shift_VCID;

				printf LOG "$VC\t$counter_value\n";
				
				$sync_found = 0;		# reset to avoid logging all frames
			}
			
			# ----- prepare output -----
			
			if ($output)
			{
				if ($only_VCDU)
				{
					$out_bytes = substr $payload_bytes, 0, $B_VCDU;		# VCDU only
				}
				elsif ($no_ann)
				{
					$out_bytes = $SM_bytes.$payload_bytes;
				}
				else		# annotated CADU
				{
					$ann_bytes = substr $annCADU_bytes, $B_CADU, $B_ann;
					$out_bytes = $SM_bytes.$payload_bytes.$ann_bytes;
				}	
			}
				
			# ----- output unless idle frames are not wanted -----
			
			if ($no_idle)
			{
				if ($VCID ne $idle_VCID)
				{
					printf ",   non-idle CADU %8d", ++$n_niFRAMEs;
		
					if ($output)
					{
						if ($per_VCID)	{ $out = $outFRAMEs[$VC]; print $out $out_bytes; }
						else			{ print outFRAMEs $out_bytes; }
					}
				}
				else
				{
					printf ",   non-idle CADU %8d", $n_niFRAMEs;
				}
			}
			else
			{
					if ($output)
					{
						if ($per_VCID)	{ $out = $outFRAMEs[$VC]; print $out $out_bytes; }
						else			{ print outFRAMEs $out_bytes; }
					}
			}
			
			printf ",   CCSDS v%1d,   SCID = (x) %2x,   VCID = (d) %2d", $CC, $SC, $VC;
			if ($keep_bad_sync) { printf ",   damaged sync marker %4d", $damaged_SM_count; }
			
			print "\r";			# print all log messages on the same line to avoid scrolling
		}
		else
		{
			$trailer_count = $bytes_read;
		}
	}
}
else		# input is VCDUs
{
	while ($bytes_read = read(inFRAMEs, $annVCDU_bytes, $B_annVCDU, 0))
	{
		if ($bytes_read == $B_annVCDU)
		{
			printf "VCDU %8d,   $bytes_read bytes", ++$n_FRAMEs;
			
			# ----- decode VCID -----
		
			$first_2bytes = substr $annVCDU_bytes, 0, 2;	# first 2 bytes
		
			$CCSDS =  $first_2bytes & $mask_CCSDS ;	
			$SCID  =  $first_2bytes & $mask_SCID ;	
			$VCID  =  $first_2bytes & $mask_VCID ;	
			
			# ----- prepare output -----
			
			if ($output)
			{
				if ($no_ann)
				{
					$out_bytes = substr $annVCDU_bytes, 0, $B_VCDU;		# VCDU only
				}
				else		# annotated VCDU
				{
					$out_bytes = $annVCDU_bytes;
				}	
			}
				
			# ----- output unless idle frames are not wanted -----
			
			$VC_count{$VCID}++;			# VCID statistics
		
			$CC = unpack "n", $CCSDS; $CC = $CC >> $shift_CCSDS; $CC++;
			$SC = unpack "n", $SCID;  $SC = $SC >> $shift_SCID;
			$VC = unpack "n", $VCID;  $VC = $VC >> $shift_VCID;

			if ($log_VCID)
			{
				$counter_bytes = "\x00".substr $annVCDU_bytes, 2, 3;	# VC counter 
				if ($out_scrambled)		# need to xor to decode VC counter
				{
					$counter_bytes = $counter_bytes ^ "\x00".$polynom_counter;
				}
				$counter_value = vec $counter_bytes, 0, 32;				# in decimal

				print LOG_VCID "$n_FRAMEs\t$VC\t$counter_value\n";
			}
			if ($no_idle)
			{
				if ($VCID ne $idle_VCID)
				{
					printf ",   non-idle VCDU %8d", ++$n_niFRAMEs;
		
					if ($output)
					{
						if ($per_VCID)	{ $out = $outFRAMEs[$VC]; print $out $out_bytes; }
						else			{ print outFRAMEs $out_bytes; }
					}
				}
				else
				{
					printf ",   non-idle VCDU %8d", $n_niFRAMEs;
				}
			}
			else
			{
					if ($output)
					{
						if ($per_VCID)	{ $out = $outFRAMEs[$VC]; print $out $out_bytes; }
						else			{ print outFRAMEs $out_bytes; }
					}
			}		
			
			printf ",   CCSDS v%1d,   SCID = (x) %2x,   VCID = (d) %2d", $CC, $SC, $VC;
			
			print "\r";			# print all log messages on the same line to avoid scrolling
		}
		else
		{
			$trailer_count = $bytes_read;
		}
	}
}

print "\n";					# final new line to leave the last message visible

print "$trailer_count trailing bytes dropped\n";		# report trailing bytes count

if ($stats)
{
	print "----- statistics -----\n";
			
	foreach $VCID (sort keys %VC_count)
	{
		printf "VCID %2d: %8d frames\n", (unpack "n", $VCID) >> $shift_VCID, $VC_count{$VCID};	
	}
}
close(inFRAMEs);

if ($output)
{
	print "----- output files -----\n";
			
	if ($per_VCID)
	{
		for ($VC = 0; $VC < $maxVC; $VC++)
		{
			close($outFRAMEs[$VC]);
			
			if (-z $f_outFRAMEs[$VC])	{ unlink $f_outFRAMEs[$VC]; }	# if empty file, delete
			else						{ print "output file:   $f_outFRAMEs[$VC]\n"; }
		}
	}
	else
	{
		close(outFRAMEs);
		print "output file:   $f_outFRAMEs\n"; 
	}
}

print "----- DONE -----\n";
print "\n";					# final new line to leave the last message visible

exit 0;
